#include <DatabaseUtils/DAQConfig/GTRunConfig.h>
#include <DatabaseUtils/DAQConfig/ConfigDataTree.h>
#include <DatabaseUtils/DAQConfig/DBColumnParams.h>
#include <DatabaseUtils/DAQConfig/DBColumnUtils.h>
#include <DatabaseUtils/DAQConfig/NamedConfigUtils.h>
#include <NovaDatabase/Table.h>
#include <NovaDAQUtilities/EnvVarCache.h>
#include <boost/algorithm/string.hpp>

namespace NOVADB = nova::database;

namespace dbutils {
namespace daqconfig {

/**
 * Creates a GT run configuration object from the specified
 * list of subsystem configuration IDs.
 *
 * @param subsystemConfigIdList The list of subsystem
 *        configuration IDs.
 * @param dcmName The name of the DCM of interest.
 */
GTRunConfig::
GTRunConfig(const std::vector<int64_t>& subsystemConfigIdList)
{
    _primarySubsystemConfigId = -1;
    if (subsystemConfigIdList.size() > 0) {
        _primarySubsystemConfigId = subsystemConfigIdList[0];
        _hasValidData = _fetchAllData(subsystemConfigIdList);
    }
    else {
        _hasValidData = false;
    }
}

/**
 * Creates a GT run configuration object from the specified
 * subsystem configuration specification.
 *
 * @param nameSpec The named configuration of interest.
 */
GTRunConfig::GTRunConfig(const SubsystemNameSpec& nameSpec)
{
    _primarySubsystemConfigId = -1;
    std::vector<int64_t> subsystemConfigIdList;
    if (NamedConfigUtils::
        getSubsystemConfigIdAncestry(nameSpec, subsystemConfigIdList)) {
        _primarySubsystemConfigId = subsystemConfigIdList[0];
        _hasValidData = _fetchAllData(subsystemConfigIdList);
    }
    else {
        _hasValidData = false;
    }
}

/**
 * Creates a GT run configuration object from the specified
 * global configuration specification.
 *
 * @param nameSpec The named configuration of interest.
 */
GTRunConfig::GTRunConfig(const GlobalNameSpec& nameSpec)
{
    _primarySubsystemConfigId = -1;
    int64_t subsystemConfigId;
    bool status = NamedConfigUtils::
        getSubsystemConfigIdFromGlobalName(nameSpec,
                                           SubsystemNameSpec::
                                           GT_SUBSYSTEM_NAME,
                                           SubsystemNameSpec::
                                           RUN_CONFIG_TYPE,
                                           subsystemConfigId);

    std::vector<int64_t> subsystemConfigIdList;
    if (status) {
        IDSpec idSpec(subsystemConfigId, IDSpec::SUBSYSTEM, IDSpec::DAQ);
        status = NamedConfigUtils::
            getConfigIdAncestry(idSpec, subsystemConfigIdList);
    }

    if (status) {
        _primarySubsystemConfigId = subsystemConfigIdList[0];
        _hasValidData = _fetchAllData(subsystemConfigIdList);
    }
    else {
        _hasValidData = false;
    }
}

/**
 * Creates a GT run configuration object from the specified
 * global configuration ID.
 *
 * @param globalConfigId The ID of the global configuration of interest.
 */
GTRunConfig::GTRunConfig(const int64_t& globalConfigId)
{
    IDSpec idSpec(globalConfigId, IDSpec::GLOBAL, IDSpec::DAQ);
    _primarySubsystemConfigId = -1;
    std::vector<int64_t> subsystemConfigIdList;
    if (NamedConfigUtils::
        getSubsystemConfigIdAncestryFromGlobalId(idSpec,
                                                 SubsystemNameSpec::
                                                 GT_SUBSYSTEM_NAME,
                                                 SubsystemNameSpec::
                                                 RUN_CONFIG_TYPE,
                                                 subsystemConfigIdList)) {
        _primarySubsystemConfigId = subsystemConfigIdList[0];
        _hasValidData = _fetchAllData(subsystemConfigIdList);
    }
    else {
        _hasValidData = false;
    }
}

bool GTRunConfig::
_fetchAllData(const std::vector<int64_t>& subsystemConfigIdList)
{
    bool status = true;
    std::string fname;

    // for the GT configuration, inheritance is not used,
    // so only the first subsystem ID is relevant
    int64_t subsystemConfigId = subsystemConfigIdList[0];

    if (status) {
        fname = "NovaDatabase/tables/DAQConfig/CalibrationTriggers.xml";
        status = _fetchTriggerData(fname, subsystemConfigId,
                                   _calibrationTriggerList);
    }
    if (status) {
        fname = "NovaDatabase/tables/DAQConfig/ExternalTriggers.xml";
        status = _fetchTriggerData(fname, subsystemConfigId,
                                   _externalTriggerList);
    }
    if (status) {
        fname = "NovaDatabase/tables/DAQConfig/ManualTriggers.xml";
        status = _fetchTriggerData(fname, subsystemConfigId,
                                   _manualTriggerList);
    }
    if (status) {
        fname = "NovaDatabase/tables/DAQConfig/SpillTriggers.xml";
        status = _fetchTriggerData(fname, subsystemConfigId,
                                   _spillTriggerList);
    }
    if (status) {
      fname = "NovaDatabase/tables/DAQConfig/DataDrivenTriggers.xml";
      status = _fetchTriggerData(fname, subsystemConfigId,
				 _dataDrivenTriggerList); 
    }
    if (status) {
      fname = "NovaDatabase/tables/DAQConfig/SNEWSTriggers.xml";
      status = _fetchTriggerData(fname, subsystemConfigId,
				 _SNEWSTriggerList); 
    }

    if (status) {
        TriggerList tmpTriggerList;
        fname = "NovaDatabase/tables/DAQConfig/TriggerOffsets.xml";
        status = _fetchTriggerData(fname, subsystemConfigId,
                                   tmpTriggerList);
        if (tmpTriggerList.size() == 1) {
            _triggerOffset = tmpTriggerList[0];
        }
    }
    if (status) {
      TriggerList tmpTriggerList;
      fname = "NovaDatabase/tables/DAQConfig/GlobalThrottle.xml";
      status = _fetchTriggerData(fname, subsystemConfigId,
				 tmpTriggerList); 
      if (tmpTriggerList.size() == 1) {
	   _globalThrottle = tmpTriggerList[0];
      }
    }
    if (status) {
      TriggerList tmpTriggerList;
      fname = "NovaDatabase/tables/DAQConfig/DDTThrottle.xml";
      status = _fetchTriggerData(fname, subsystemConfigId,
				 tmpTriggerList);
      if (tmpTriggerList.size() == 1) {
	   _DDTThrottle = tmpTriggerList[0];
      } 
    }


    return status;
}

bool GTRunConfig::
_fetchTriggerData(const std::string& tableDefinitionFileName,
                  const int64_t& subsystemConfigId,
                  TriggerList& triggerList)
{
    boost::shared_ptr<NOVADB::Table> dbt;
    dbt.reset(new NOVADB::Table(tableDefinitionFileName));
    std::string detName = 
      novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(detName);
    std::string dbUser =
      novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {dbt->SetUser(dbUser);}
    if (! dbt->GetConnection()) {return false;}
    if (! dbt->ExistsInDB()) {return false;}

    dbt->Clear();
    dbt->SetValidityRange(NamedConfigUtils::SUBSYSCFGID_COLUMN_NAME,
                          subsystemConfigId);
    if (! dbt->LoadFromDB()) {return false;}

    std::vector<std::string> colNames = dbt->GetColNames();
    NOVADB::Row* rowPtr;
    NOVADB::Column* colPtr;
    for (int rowIdx = 0; rowIdx < dbt->NRow(); ++rowIdx) {
        boost::shared_ptr<SimpleParameterSet>
            trigParamSet(new SimpleParameterSet());
        rowPtr = dbt->GetRow(rowIdx);

        for (uint32_t colIdx = 0; colIdx < colNames.size(); ++colIdx) {
            std::string colName = boost::to_lower_copy(colNames[colIdx]);
            if (DBColumnUtils::isInfrastructureColumn(colName)) {continue;}
            if (DBColumnUtils::isSubsystemConfigReferenceColumn(colName)) {
                continue;
            }

            std::string valueString;
            colPtr = rowPtr->Col(colName);
            if (! colPtr->Get(valueString)) {
                return false;
            }
            trigParamSet->put(colName, valueString);
        }

        triggerList.push_back(trigParamSet);
    }

    return true;
}

};  // end of namespace daqconfig
};  // end of namespace dbutils
